#ifndef IMAP_URLAUTH_WORKER_COMMON_H
#define IMAP_URLAUTH_WORKER_COMMON_H

#define IMAP_URLAUTH_WORKER_SOCKET "imap-urlauth-worker"

#define IMAP_URLAUTH_WORKER_PROTOCOL_MAJOR_VERSION 2
#define IMAP_URLAUTH_WORKER_PROTOCOL_MINOR_VERSION 0

#endif
