static const char *stats_metric_defaults =
"group @metric_defaults proxy {\n"
"metric auth_successes {\n"
"  filter = event=auth_request_finished AND success=yes\n"
"}\n"
"metric auth_failures {\n"
"  filter = event=auth_request_finished AND NOT success=yes\n"
"}\n"
"metric login_aborted {\n"
"  filter = event=login_aborted\n"
"  group_by reason {\n"
"  }\n"
"}\n"
"metric proxy_session_established {\n"
"  filter = event=proxy_session_established\n"
"}\n"
"}\n"

"group @metric_defaults backend {\n"
"metric auth_successes {\n"
"  filter = event=auth_request_finished AND success=yes\n"
"}\n"
"metric auth_failures {\n"
"  filter = event=auth_request_finished AND NOT success=yes\n"
"}\n"
"metric imap_commands {\n"
"  filter = event=imap_command_finished\n"
"  group_by tagged_reply_state {\n"
"  }\n"
"}\n"
"metric mail_deliveries {\n"
"  filter = event=mail_delivery_finished\n"
"}\n"
"metric mail_submissions {\n"
"  filter = event=smtp_submit_finished\n"
"}\n"
"metric mail_user_session_finished {\n"
"   filter = event=mail_user_session_finished\n"
"   fields = rss utime\n"
"}\n"
"}\n";

static const char *mailbox_defaults =
"group @mailbox_defaults english {\n"
"  namespace inbox {\n"
"    inbox = yes\n"
"    mailbox Drafts {\n"
"      auto = subscribe\n"
"      special_use = \\Drafts\n"
"    }\n"
"    mailbox Junk {\n"
"      auto = subscribe\n"
"      special_use = \\Junk\n"
"    }\n"
"    mailbox Sent {\n"
"      auto = subscribe\n"
"      special_use = \\Sent\n"
"    }\n"
"    mailbox Trash {\n"
"      auto = subscribe\n"
"      special_use = \\Trash\n"
"    }\n"
"  }\n"
"}\n";
