#ifndef ALL_SETTINGS_H
#define ALL_SETTINGS_H

extern const struct setting_parser_info *const *all_infos;
extern const struct setting_parser_info *all_default_infos[];
extern const struct config_service *config_all_services;

#endif
