static const struct setting_history_rename settings_history_core_renames[] = {
};
static const struct setting_history_default settings_history_core_defaults[] = {
  { "service/lmtp/service_extra_groups", "", "2.4.4" },
  { "push_notification_ox/http_client_request_max_attempts", "1", "2.4.4" },
  { "push_notification_ox/http_client_request_timeout", "1min", "2.4.4" },
  { "service/imap/unix_listener/imap-master/user", "", "2.4.3" },
  { "service/anvil/unix_listener/anvil-auth-penalty/mode", "0600", "2.4.3" },
  { "service/anvil/unix_listener/anvil-auth-penalty/group", "", "2.4.3" },
  { "@mailbox_defaults/english/namespace/inbox/mailbox/Drafts/mailbox_auto", "no", "2.4.3" },
  { "@mailbox_defaults/english/namespace/inbox/mailbox/Junk/mailbox_auto", "no", "2.4.3" },
  { "@mailbox_defaults/english/namespace/inbox/mailbox/Sent/mailbox_auto", "no", "2.4.3" },
  { "@mailbox_defaults/english/namespace/inbox/mailbox/Trash/mailbox_auto", "no", "2.4.3" },
  { "service/doveadm/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/imap/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/imap-urlauth-worker/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/indexer-worker/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/lmtp/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/pop3/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "service/submission/service_extra_groups", "$SET:default_internal_group", "2.4.3" },
  { "mail_access_groups", "", "2.4.3" },
  { "lazy_expunge_only_last_instance", "no", "2.4.3" },
  { "passdb_ldap/passdb_default_password_scheme", "CRYPT", "2.4.3" },
  { "passdb_passwd/passdb_default_password_scheme", "CRYPT", "2.4.3" },
  { "passdb_default_password_scheme", "PLAIN", "2.4.3" },
  { "maildir/mail_attachment_detection_options", "add-flags content-type=!application/signature", "2.4.3" },
  { "maildir/mail_attachment_detection_options", "", "2.4.2" },
  { "mail_attachment_detection_options", "", "2.4.2" },
  { "@metric_defaults/proxy/metric", "auth_successes,auth_failures,login_aborted", "2.4.2" },
  { "mailbox_directory_name_legacy", "yes", "2.4.2" },
  { "service/anvil/unix_listener/anvil/unix_listener_mode", "0600", "2.4.2" },
  { "service/anvil/unix_listener/anvil/unix_listener_group", "", "2.4.2" },
  { "service/lmtp/service_restart_request_count", "unlimited", "2.4.1" },
  { "service/auth/unix_listener/auth-userdb/unix_listener_group", "", "2.4.1" },
  { "mail_cache_fields", "flags", "2.4.1" },
  { "lmtp_user_concurrency_limit", "0", "2.4.1" },
};
