/* Copyright (c) the Dovecot authors, based on code by Michael Slusarz.
 * See the included COPYING file */

#ifndef DOVEADM_DUMP_FLATCURVE_H
#define DOVEADM_DUMP_FLATCURVE_H

void doveadm_dump_flatcurve_init(void);

#endif
