/* This file may be multiply-included, with different definitions of
   'TEST()' macro. This is sometimes called "the X trick" (as the
   macro is often imaginatively called X(). */

TEST(test_aqueue)
TEST(test_array)
FATAL(fatal_array)
TEST(test_backtrace)
TEST(test_base32)
TEST(test_base64)
TEST(test_bits)
TEST(test_bsearch_insert_pos)
TEST(test_buffer)
TEST(test_buffer_append_full)
FATAL(fatal_buffer)
TEST(test_byteorder)
TEST(test_connection)
TEST(test_crc32)
TEST(test_data_stack)
FATAL(fatal_data_stack)
TEST(test_env_util)
FATAL(fatal_env_util)
TEST(test_event_category_register)
FATAL(fatal_event_category_register)
TEST(test_lib_event)
FATAL(fatal_lib_event)
TEST(test_event_filter)
TEST(test_event_filter_expr)
TEST(test_event_filter_merge)
TEST(test_event_filter_parser)
TEST(test_event_flatten)
TEST(test_event_log)
TEST(test_failures)
TEST(test_file_cache)
TEST(test_file_create_locked)
TEST(test_guid)
TEST(test_hash)
TEST(test_hash_format)
TEST(test_hash_method)
TEST(test_hmac)
TEST(test_hex_binary)
FATAL(fatal_i_close)
TEST(test_imem)
TEST(test_ioloop)
TEST(test_iso8601_date)
TEST(test_iostream_pump)
TEST(test_iostream_proxy)
TEST(test_iostream_temp)
TEST(test_istream)
TEST(test_istream_base64_decoder)
TEST(test_istream_base64_encoder)
TEST(test_istream_chain)
TEST(test_istream_concat)
TEST(test_istream_crlf)
TEST(test_istream_failure_at)
TEST(test_istream_multiplex)
TEST(test_istream_noop)
TEST(test_istream_seekable)
TEST(test_istream_sized)
TEST(test_istream_tee)
TEST(test_istream_try)
TEST(test_istream_unix)
TEST(test_lib_event)
TEST(test_lib_signals)
TEST(test_llist)
TEST(test_log_throttle)
TEST(test_macros)
TEST(test_malloc_overflow)
FATAL(fatal_malloc_overflow)
TEST(test_memarea)
TEST(test_mempool)
FATAL(fatal_mempool)
TEST(test_mempool_alloconly)
FATAL(fatal_mempool_alloconly)
TEST(test_mempool_allocfree)
FATAL(fatal_mempool_allocfree)
TEST(test_net)
TEST(test_numpack)
TEST(test_ostream_buffer)
TEST(test_ostream_failure_at)
TEST(test_ostream_file)
FATAL(fatal_ostream_file)
TEST(test_ostream_multiplex)
TEST(test_multiplex)
TEST(test_path_util)
TEST(test_pkcs5_pbkdf2)
TEST(test_primes)
TEST(test_printf_format_fix)
FATAL(fatal_printf_format_fix)
TEST(test_priorityq)
TEST(test_punycode)
TEST(test_random)
FATAL(fatal_random)
TEST(test_seq_range_array)
FATAL(fatal_seq_range_array)
TEST(test_seq_set_builder)
TEST(test_stats_dist)
TEST(test_str)
TEST(test_strescape)
TEST(test_strfuncs)
FATAL(fatal_strfuncs)
TEST(test_strnum)
TEST(test_str_find)
TEST(test_str_parse)
TEST(test_str_sanitize)
TEST(test_str_table)
TEST(test_time_util)
TEST(test_unichar)
TEST(test_unicode_break)
TEST(test_unicode_data)
TEST(test_unicode_nf)
TEST(test_unicode_casemap)
TEST(test_uri)
TEST(test_utc_mktime)
TEST(test_version)
TEST(test_wildcard_match)
